#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import argparse
import logging
import os
import sys
from collections import defaultdict
from os import PathLike
from pathlib import Path

from majormode.perseus.constant.logging import LOGGING_LEVEL_LITERAL_STRINGS
from majormode.perseus.constant.logging import LoggingLevelLiteral
from majormode.perseus.utils import cast
from majormode.perseus.utils.logging import DEFAULT_LOGGING_FORMATTER
from majormode.perseus.utils.logging import cast_string_to_logging_level
from majormode.perseus.utils.logging import set_up_logger

from bootloader.utils.plasticscm import reporting
from bootloader.utils.plasticscm.model import Changeset, UserStats


def cast_string_to_path(path: str) -> PathLike:
    return Path(path)


def parse_arguments() -> argparse.Namespace:
    """
    Convert argument strings to objects and assign them as attributes of
    the namespace.


    @return: An instance `Namespace` corresponding to the populated
        namespace.
    """
    parser = argparse.ArgumentParser(description="Plastic SCM repository activities reporting")

    parser.add_argument(
        '--end-time',
        dest='end_time',
        metavar='ISO 8601 Timestamp',
        required=False,
        type=cast.string_to_timestamp,
        help="specify the latest date of changesets to return.  This date is "
             "exclusive, so changesets that were made at this date are not "
             "returned."
    )

    parser.add_argument(
        '--logging-level',
        dest='logging_level',
        metavar='LEVEL',
        required=False,
        default=str(LoggingLevelLiteral.info),
        type=cast_string_to_logging_level,
        help=f"specify the logging level ({', '.join(LOGGING_LEVEL_LITERAL_STRINGS)})"
    )

    parser.add_argument(
        '--server',
        dest='plasticscm_server_address',
        metavar='NAME',
        required=False,
        help="specify the Plastic SCM server to connect to"
    )

    parser.add_argument(
        '-o, --output-file',
        dest='stats_output_file_path_name',
        metavar='PATH',
        required=False,
        type=cast_string_to_path,
        help="specify the path and name of the file to write in the activity "
             "statistics of the Plastic SCM repositories"
    )

    parser.add_argument(
        '--start-time',
        dest='start_time',
        metavar='ISO 8601 Timestamp',
        required=False,
        type=cast.string_to_timestamp,
        help="specify the earliest date of changesets to return.  This date "
             "is inclusive, so changesets that were made at this date are "
             "returned."
    )

    return parser.parse_args()


def run():
    arguments = parse_arguments()

    set_up_logger(
        logging_formatter=DEFAULT_LOGGING_FORMATTER,
        logging_level=arguments.logging_level
    )

    plastic_scm_version = reporting.get_cm_version()
    logging.debug(f"Plastic SCM command line v{plastic_scm_version}")

    # Determine the list of Plastic SCM repositories available on the
    # specified server.
    available_repositories = reporting.fetch_available_repositories(
        server_address=arguments.plasticscm_server_address
    )

    # Calculate the activity statistics of the Plastic SCM repositories per
    # user.
    users_stats: dict[str, UserStats] = {}

    for repository_name in available_repositories:
        changesets = reporting.fetch_changeset_history(
            repository_name,
            include_details=True,
            end_date=arguments.end_time,
            start_date=arguments.start_time
        )

        for changeset in changesets:
            user_stats = users_stats.get(changeset.owner)
            if user_stats is None:
                user_stats = UserStats(changeset.owner)
                users_stats[changeset.owner] = user_stats
            user_stats.add_changeset(changeset)

    # Aggregate all the users' activity statistics.
    stats_report = [
        user_stats.generate_stats_report()
        for user_stats in users_stats.values()
    ]

    # Write the stats report into the specified file if any defined,
    # otherwise print the stats report onto the standard output.
    stringified_stats_report = os.linesep.join(stats_report)
    if arguments.stats_output_file_path_name is not None:
        with open(arguments.stats_output_file_path_name, 'wt') as fd:
            fd.writelines(stringified_stats_report)
    else:
        print(stringified_stats_report)





# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from __future__ import annotations

import datetime
import os
import re
from collections import defaultdict
from os import PathLike
from pathlib import Path
from typing import Any

from majormode.perseus.model.date import ISO8601DateTime
from majormode.perseus.utils import cast


class Changeset:
    """
    A changeset describes the exact differences between two successive
    versions in the version control system's repository of changes.
    Changesets are typically treated as an atomic unit, an indivisible set,
    by version control systems.
    """
    CSV_FIELD_INDEX_OWNER = 0
    CSV_FIELD_INDEX_DATE = 1
    CSV_FIELD_INDEX_CHANGESET_ID = 2

    # REGEX_PATTERN_PLASTIC_SCM_TIMESTAMP = r'^(?P<month>[1-9]|1[0-9])\/(?P<day>[1-9]|[1-2][0-9]|3[01])\/(?P<year>2[0-9]{3})\s(?P<hour>[0-9]|1[0-2]):(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9])(?P<meridiem>AM|PM)$'
    REGEX_PATTERN_PLASTIC_SCM_TIMESTAMP = r'^(?P<month>\d{1,2})\/(?P<day>\d{1,2})\/(?P<year>\d{4})\s(?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})(?P<meridiem>AM|PM)?$'
    REGEX_PLASTIC_SCM_TIMESTAMP = re.compile(REGEX_PATTERN_PLASTIC_SCM_TIMESTAMP)

    # @classmethod
    # def __cast_stringified_timestamp_to_iso8601(cls, timestamp: str) -> ISO8601DateTime:
    #     """
    #     10/18/2023
    #
    #     :param timestamp:
    #
    #     :return:
    #     """
    #     match = cls.REGEX_PLASTIC_SCM_TIMESTAMP.match(timestamp)
    #     if match is None:
    #         raise ValueError(f"Invalid timestamp '{timestamp}'")
    #
    #     day = match.group('day')
    #     month = match.group('month')
    #     year = match.group('year')
    #
    #     hours = int(match.group('hour'))
    #     minutes = match.group('minute')
    #     seconds = match.group('second')
    #
    #     meridiem = match.group('meridiem')
    #
    #     if meridiem == 'AM' and hours == 12:
    #         hours = '0'
    #     elif meridiem == 'PM' and hours < 12:
    #         hours = str(int(hours) + 12)
    #
    #     stringified_iso_8601_timestamp = f'{year}-{month:>02}-{day:>02}T{hours:>02}:{minutes}:{seconds}'
    #
    #     return cast.string_to_timestamp(stringified_iso_8601_timestamp)

    def __init__(
            self,
            repository_name: str,
            owner: str,
            timestamp: ISO8601DateTime,
            changeset_id: int):
        self.__repository_name = repository_name
        self.__owner = owner
        self.__timestamp = timestamp
        self.__changeset_id = changeset_id
        self.__details = None

    @property
    def changeset_id(self) -> int:
        return self.__changeset_id

    @property
    def date(self) -> date:
        return datetime.date(
            day=self.__timestamp.day,
            month=self.__timestamp.month,
            year=self.__timestamp.year
        )

    @property
    def details(self) -> list[ChangesetDetail]:
        return self.__details

    @details.setter
    def details(self, details: list[ChangesetDetail]) -> None:
        if self.__details is not None:
            raise ValueError(f"The changeset {self.__changeset_id} has already details defined")
        self.__details = details

    @classmethod
    def from_csv(cls, repository_name: str, payload: Any) -> Changeset or None:
        return payload and Changeset(
            repository_name,
            payload[cls.CSV_FIELD_INDEX_OWNER],
            # cls.__cast_stringified_timestamp_to_iso8601(payload[cls.CSV_FIELD_INDEX_DATE]),
            cast.string_to_timestamp(payload[cls.CSV_FIELD_INDEX_DATE]),
            int(payload[cls.CSV_FIELD_INDEX_CHANGESET_ID])
        )

    @property
    def owner(self) -> str:
        return self.__owner

    @property
    def project_name(self) -> str:
        project_name, organization_name, server_address = self.__repository_name.split('@')
        return project_name

    @property
    def repository_name(self) -> str:
        return self.__repository_name

    @property
    def timestamp(self) -> ISO8601DateTime:
        return self.__timestamp


class ChangesetDetail:
    CSV_FIELD_INDEX_FILE_PATH = 0
    CSV_FIELD_INDEX_FILE_TYPE = 1
    CSV_FIELD_INDEX_FILE_SIZE = 2

    def __init__(
            self,
            file_path: PathLike,
            file_type: str,
            file_size: int,
            lines_deleted: int or None= None,
            lines_inserted: int or None = None
    ):
        self.__file_path = file_path
        self.__file_type = file_type
        self.__file_size = file_size
        self.__lines_deleted = lines_deleted
        self.__lines_inserted = lines_inserted

    @property
    def file_path(self) -> PathLike:
        return self.__file_path

    @property
    def file_size(self) -> int:
        return self.__file_size

    @property
    def file_type(self) -> str:
        return self.__file_type

    @classmethod
    def from_csv(cls, payload: Any) -> ChangesetDetail:
        return payload and ChangesetDetail(
            Path(payload[cls.CSV_FIELD_INDEX_FILE_PATH]),
            payload[cls.CSV_FIELD_INDEX_FILE_TYPE],
            int(payload[cls.CSV_FIELD_INDEX_FILE_SIZE])
        )

    @property
    def lines_deleted(self) -> int or None:
        return self.__lines_deleted

    @lines_deleted.setter
    def lines_deleted(self, lines_deleted: int):
        if self.__lines_deleted is not None:
            raise ValueError(f"The number of deleted lines has been already defined ({self.lines_deleted})")
        self.__lines_deleted = lines_deleted

    @property
    def lines_changed(self) -> int or None:
        return None if self.__lines_inserted is None or self.__lines_deleted is None \
            else self.__lines_inserted + self.__lines_deleted

    @property
    def lines_inserted(self) -> int or None:
        return self.__lines_inserted

    @lines_inserted.setter
    def lines_inserted(self, lines_inserted: int):
        if self.__lines_inserted is not None:
            raise ValueError(f"The number of inserted lines has been already defined ({self.__lines_inserted})")
        self.__lines_inserted = lines_inserted


class UserStats:
    class DateStats:
        def __init__(self, date: datetime.date):
            self.date = date
            self.files_changed = 0
            self.lines_changed = 0
            self.unknown_changed = 0

    def __init__(self, username: str):
        self.__username = username
        self.__repository_changesets = defaultdict(list[Changeset])

    def add_changeset(self, changeset: Changeset):
        self.__repository_changesets[changeset.repository_name].append(changeset)

    @property
    def username(self) -> str:
        return self.__username

    def generate_stats_report(self) -> str:
        stats_report = [f"{self.__username}:"]

        for changesets in self.__repository_changesets.values():
            project_name = changesets[0].project_name

            dates_stats: dict[datetime.date, UserStats.DateStats] = {}

            for changeset in changesets:
                date_stats = dates_stats.get(changeset.date)
                if date_stats is None:
                    dates_stats[changeset.date] = date_stats = UserStats.DateStats(changeset.date)

                date_stats.files_changed += len(changeset.details)

                for detail in changeset.details:
                    if detail.file_type == 'txt':
                        if detail.lines_changed is None:
                            date_stats.unknown_changed += 1
                        else:
                            date_stats.lines_changed += detail.lines_changed

            for date_stats in dates_stats.values():
                stats = f"- [{project_name}][{date_stats.date}] " \
                    f"{len(changesets)} changesets, " \
                    f"{date_stats.files_changed} files changed, " \
                    f'{date_stats.lines_changed} lines changed'

                if date_stats.unknown_changed > 0:
                    stats += f" ({date_stats.unknown_changed} unknown changes)"

                stats_report.append(stats)

        return os.linesep.join(stats_report)




# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import logging
import re
import subprocess
from os import PathLike
from tempfile import TemporaryFile, NamedTemporaryFile

from majormode.perseus.model.date import ISO8601DateTime

from bootloader.utils.plasticscm.model import Changeset
from bootloader.utils.plasticscm.model import ChangesetDetail


REGEX_PATTERN_DIFFSTAT_OUTPUT = r'\s*\d+.*changed,\s+(?P<lines_inserted>\d*)\s+insertion.*(?P<lines_deleted>\d+)\s+deletion.*'
REGEX_DIFFSTAT_OUTPUT = re.compile(REGEX_PATTERN_DIFFSTAT_OUTPUT)


def convert_to_plastics_scm_timestamp(timestamp: ISO8601DateTime) -> str:
    """

    :param timestamp:
    :return:
    """
    plastic_scm_timestamp = f'{timestamp.year:>04}-{timestamp.month:>02}-{timestamp.day:>02}'

    if hasattr(plastic_scm_timestamp, 'hour'):
        plastic_scm_timestamp += f' {timestamp.hour:>02}:{timestamp.minute:>02}:{timestamp.second:>02}'
    else:
        plastic_scm_timestamp += ' 00:00:00'

    return plastic_scm_timestamp


def get_cm_version() -> str:
    command_line = ['cm', 'version']

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    return result.stdout.strip()


def fetch_available_repositories(server_address: str = None) -> list[str]:
    """
    Return the list of available repositories.

    The Plastic SCM server address is also known as the "repository server
    spec" (``repserverspec``).  It can be represented with the string
    ``repserver:name:port``. For examples:

    ```text
    repserver:skull:8084
    skull:8084
    ```

    Plastic SCM call it "repository server spec", instead of just "server
    spec" for historical reasons.  Long ago, they had separate workspace
    and repository servers, and the naming survived.


    :param server_address The hostname and port of the Plastic SCM server
        to connect to and return the list of available repositories.  When
        not defined, the default server is used, i.e., the server
        configured in the configuration wizard and is written in the local
        ``client.conf`` file.


    :return: An array of the names of available repositories.
    """
    command_line = ['cm', 'repository', 'list']
    if server_address is not None:
        command_line.append(server_address)

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    return result.stdout.strip().split('\n')


def fetch_changeset_details(
        changeset: Changeset,
        include_file_diff_stats: bool = False
) -> list[ChangesetDetail]:
    """
    Return the list of files and lines that have been changed in the
    specified changeset.


    :param changeset: The changeset to return detailed information.

    :param include_file_diff_stats: Indicate whether to include statistics
        about file changes.


    :return: The list of files that have been changed in the changeset.
    """
    command_line = ['cm', 'find', 'revisions']
    # command_line.append(f"where changeset={changeset.changeset_id} and type!='dir'")
    command_line.append(f"where changeset={changeset.changeset_id} and type='txt'")
    command_line.extend(['on', 'repository', f"'{changeset.repository_name}'"])
    command_line.append('--format={item};{type};{size}')
    command_line.append('--nototal')

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    changeset_detail_records = result.stdout.strip().split('\n')

    changeset_details = [
        ChangesetDetail.from_csv(record.split(';'))
        for record in changeset_detail_records
        if record != ''
    ]

    if include_file_diff_stats:
        for detail in changeset_details:
            if detail.file_type == 'txt':
                lines_inserted, lines_deleted = fetch_file_diff_stats(detail.file_path, changeset)
                detail.lines_inserted = lines_inserted
                detail.lines_deleted = lines_deleted

    return changeset_details


# Fetch commit history
def fetch_changeset_history(
        repository_name: str,
        end_date: ISO8601DateTime = None,
        include_details: bool = False,
        start_date: ISO8601DateTime = None,
) -> list[Changeset]:
    """
    Return a history of changesets from the specified repository.


    :param repository_name: The name of the Plastic SCM repository to
        return the changeset history.

    :param end_date: The latest date of changesets to return.  This date
        is exclusive, so changesets that were made at this date are not
        returned.

    :param include_details: Indicate to return the details of each
        changeset.

    :param start_date: The earliest date of changesets to return.  This
        date is inclusive, so changesets that were made at this date are
        returned.


    :return: A list of tuple ``(owner, date, changeset_id)``.
    """
    command_line = ['cm', 'find', 'changeset']

    if start_date or end_date:
        filters = []
        if start_date:
            filters.append(f"date >= '{convert_to_plastics_scm_timestamp(start_date)}'")
        if end_date:
            filters.append(f"date < '{convert_to_plastics_scm_timestamp(end_date)}'")

        where_section = f"where {' and '.join(filters)}"
        command_line.append(where_section)

    command_line.extend(['on', 'repository', f"'{repository_name}'"])
    command_line.append('--format={owner};{date};{changesetid}')

    # https://learn.microsoft.com/en-us/system-center/orchestrator/standard-activities/format-date-time?view=sc-orch-2022
    command_line.append('--dateformat=yyyy-MM-dd HH:mm:ss')
    command_line.append('--nototal')

    shell_command_line = ' '.join(command_line)
    logging.debug(f"Running the command line '{shell_command_line}'...")

    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    changeset_records = result.stdout.strip().split('\n')

    changesets = [
        Changeset.from_csv(
            repository_name,
            record.split(';')
        )
        for record in changeset_records
        if record != ''
    ]

    if include_details:
        for changeset in changesets:
            changeset.details = fetch_changeset_details(changeset, include_file_diff_stats=True)

    return changesets


def fetch_file_diff_stats(file_path_name: PathLike, changeset: Changeset):
    command_line = ['cm', 'diff']
    command_line.append(f'serverpath:{file_path_name}#cs:{changeset.changeset_id}@{changeset.repository_name}')

    shell_command_line = ' '.join(command_line)
    logging.debug(f"Running the command line '{shell_command_line}'...")

    with TemporaryFile() as fd:
        # print(fd.name)
        diff_process = subprocess.run(command_line, stdout=fd)
        if diff_process.returncode > 0:
            # raise Exception(diff_process.stderr)
            return None, None

        fd.seek(0)

        diffstat_process = subprocess.run(
            ['diffstat'],
            stdin=fd,
            stdout=subprocess.PIPE
        )
        if diffstat_process.returncode > 0:
            raise Exception(diff_process.stderr)

        diff_statistics = diffstat_process.stdout.decode('utf-8').strip()
        match = REGEX_DIFFSTAT_OUTPUT.search(diff_statistics)
        if match is None:
            # raise Exception(f"Invalid 'diffstat' result:\n{diff_process.stdout}")
            return None, None

        return (
            int(match.group('lines_inserted')),
            int(match.group('lines_deleted'))
        )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bootloader',
 'bootloader.cli',
 'bootloader.utils',
 'bootloader.utils.plasticscm']

package_data = \
{'': ['*']}

install_requires = \
['perseus-core-library>=1.19,<2.0']

entry_points = \
{'console_scripts': ['cmstats = bootloader.cli.cmstats:run']}

setup_kwargs = {
    'name': 'plasticscm-statistics',
    'version': '0.0.9',
    'description': 'Python command line library to generate Plastic SCM activities reports',
    'long_description': '# Plastic SCM Repository Activities Reporting\nPython command line library to generate Plastic SCM repository activities reports.\n',
    'author': 'Daniel CAUNE',
    'author_email': 'daniel@bootloader.studio',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bootloader-studio/cli-plasticscm-statistics.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
